@@ -79,7 +79,6 @@ class Lexer:
         "DOUBLE_DOT": r"\.\.",
         "DOUBLE_PIPE": r"\|\|",
         "ASSIGN": r"=",
-        "ROOT": r"\$",
         "LPAREN": r"\(",
         "RPAREN": r"\)",
         "SINGLE_QUOTE_STRING": r"'",
@@ -88,6 +87,8 @@ class Lexer:
         "COMMA": r",",
         "PIPE": r"\|",
         "LBRACKET": r"\[",
+        "EXCLAIM": r"!",
+        "QUESTION": r"\?",
     }
 
     NUMBERS: dict[str, str] = {
@@ -136,6 +137,8 @@ class Lexer:
         "COLON": TokenType.COLON,
         "COMMA": TokenType.COMMA,
         "PIPE": TokenType.PIPE,
+        "EXCLAIM": TokenType.EXCLAIM,
+        "QUESTION": TokenType.QUESTION,
     }
 
     MARKUP: dict[str, str] = {
@@ -298,6 +301,10 @@ class Lexer:
                 self.error("unexpected end of path")
 
             if c == ".":
+                if self.peek() == ".":  # probably a range expression delimiter
+                    self.backup()
+                    return
+
                 self.ignore()
                 self.ignore_whitespace()
                 if match := self.RE_PROPERTY.match(self.source, self.pos):
@@ -305,6 +312,8 @@ class Lexer:
                     self.pos += match.end() - match.start()
                     self.start = self.pos
                     self.path_stack[-1].stop = self.pos
+                else:
+                    self.error("expected a property name")
 
             elif c == "]":
                 if len(self.path_stack) == 1:
@@ -335,6 +344,7 @@ class Lexer:
                         )
                     self.next()
                     self.ignore()  # skip closing quote
+                    self.ignore_whitespace()
 
                     if self.next() != "]":
                         self.backup()
@@ -347,6 +357,7 @@ class Lexer:
                     self.path_stack[-1].path.append(int(match.group()))
                     self.pos += match.end() - match.start()
                     self.start = self.pos
+                    self.ignore_whitespace()
 
                     if self.next() != "]":
                         self.backup()
@@ -626,36 +637,10 @@ class Lexer:
 
         if kind == "SINGLE_QUOTE_STRING":
             self.ignore()
-            # self.accept_string(quote="'")
-            # expression.append(
-            #     Token(
-            #         type_=TokenType.SINGLE_QUOTE_STRING,
-            #         value=self.source[self.start : self.pos],
-            #         index=self.start,
-            #         source=self.source,
-            #     )
-            # )
-            # self.start = self.pos
-            # assert self.next() == "'"
-            # self.ignore()
             self.accept_template_string(quote="'", expression=expression)
-
         elif kind == "DOUBLE_QUOTE_STRING":
             self.ignore()
-            # self.accept_string(quote='"')
-            # expression.append(
-            #     Token(
-            #         type_=TokenType.DOUBLE_QUOTE_STRING,
-            #         value=self.source[self.start : self.pos],
-            #         index=self.start,
-            #         source=self.source,
-            #     )
-            # )
-            # self.start = self.pos
-            # assert self.next() == '"'
-            # self.ignore()
             self.accept_template_string(quote='"', expression=expression)
-
         elif kind == "LBRACKET":
             self.backup()
             self.accept_path()
@@ -706,7 +691,7 @@ class Lexer:
                 self.accept_range()
                 self.in_range = False
         else:
-            msg = f"unexpected token {self.source[self.start:self.pos]!r}"
+            msg = f"unexpected token {self.source[self.start : self.pos]!r}"
             raise LiquidSyntaxError(
                 msg,
                 token=ErrorToken(
@@ -727,7 +712,7 @@ class Lexer:
         if self.pos != self.start:
             msg = (
                 "must emit or ignore before consuming whitespace "
-                f"({self.source[self.start: self.pos]!r}:{self.pos})"
+                f"({self.source[self.start : self.pos]!r}:{self.pos})"
             )
             raise Exception(msg)
 
@@ -742,7 +727,7 @@ class Lexer:
         if self.pos != self.start:
             msg = (
                 "must emit or ignore before consuming whitespace "
-                f"({self.source[self.start: self.pos]!r}:{self.pos})"
+                f"({self.source[self.start : self.pos]!r}:{self.pos})"
             )
             raise Exception(msg)
 
@@ -758,7 +743,7 @@ class Lexer:
         if self.pos != self.start:
             msg = (
                 "must emit or ignore before consuming whitespace "
-                f"({self.source[self.start: self.pos]!r}:{self.pos})"
+                f"({self.source[self.start : self.pos]!r}:{self.pos})"
             )
             raise Exception(msg)
 
@@ -803,9 +788,9 @@ class Lexer:
             match = self.MARKUP_RULES.match(self.source, pos=self.pos)
 
             if not match:
-                assert self.pos == len(
-                    self.source
-                ), f"{self.pos}:{self.source[self.pos: 10]!r}.."
+                assert self.pos == len(self.source), (
+                    f"{self.pos}:{self.source[self.pos : 10]!r}.."
+                )
                 return None
 
             kind = match.lastgroup
